/*
	pcfxtools -- a set of tools for NEC PC-FX development and reverse-engineering
	pcfx-cdlink; .cue and .bin generator.

Copyright (C) 2011		Alex Marshall "trap15" <trap15@raidenii.net>
Copyright (C) 2007		Ryphecha / Mednafen

# This code is licensed to you under the terms of the MIT license;
# see file LICENSE or http://www.opensource.org/licenses/mit-license.php
*/

// Modified by Matej Horvat for use with Open Watcom on DOS/Win32

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
//#if WIN32
//#include <winsock.h>
//#endif

uint32_t le32(uint32_t i)
{
/*
// DOS and Win32 only run on little endian architectures anyway
	i = ntohl(i);
	i = ((i & 0xFF000000) >> 24) |
	    ((i & 0x00FF0000) >>  8) |
	    ((i & 0x0000FF00) <<  8) |
	    ((i & 0x000000FF) << 24);
*/
	return i;
}

uint16_t le16(uint16_t i)
{
/*
// DOS and Win32 only run on little endian architectures anyway
	i = ntohs(i);
	i = ((i & 0xFF00) >> 8) |
	    ((i & 0x00FF) << 8);
*/
	return i;
}

typedef struct {
	char		header[0x10];
	char		unk[0x7F0];
	char		title[0x20];
	uint32_t	sect_off;
	uint32_t	sect_count;
	uint32_t	prog_off;
	uint32_t	prog_point;
	char		maker_id[4];
	char		maker_name[60];
	uint32_t	volume_no;
	uint16_t	version;
	uint16_t	country;
	char		date[8];
	char		pad[0x380];
	char		udata[0x400];
} tyBootHeader;

tyBootHeader BootHeader = {
	"PC-FX:Hu_CD-ROM ",
	{
#include "boot.h"
	},
	"",
	0,
	0,
	0,
	0,
	"N/A",
	"pcfx-cdlink",
	0,
	0x0100,
	1,
	"20XX0131",
	{ 0, },
	{ 0, },
};

int main(int argc, char *argv[])
{
	FILE* fp;
	int keep_going = 1;
	char tmpbuf[256];
	char binname[256] = "\0";
	int i;
	int binblocks = 0;
	FILE *in_fp, *out_fp;
	struct stat stat_buf;
	char *obinname;
	uint32_t padcnt = 0;
	uint32_t prog_offset = 0x8000;
	uint32_t prog_point = 0x8000;
	char *cuename;
	uint32_t sector_offset = 2;
	int t;
	uint8_t* data;
	int32_t sh_size;
	uint32_t sector_count;

	if(argc < 3) {
		printf("Usage: %s input.txt outfiles\n", argv[0]);
		return EXIT_FAILURE;
	}
	fp = fopen(argv[1], "r");
	if(fp == NULL) {
		perror("Opening input list");
		return EXIT_FAILURE;
	}
	while(keep_going) {
		if(fgets(tmpbuf, 255, fp) == NULL)
			break;
		if(feof(fp))
			keep_going = 0;
		if(memcmp(tmpbuf, "binary ", 7) == 0) {
			snprintf(binname, (255 > (strlen(tmpbuf) - 7)) ? strlen(tmpbuf) - 7 : 255, "%s", tmpbuf + 7);
		}else if(memcmp(tmpbuf, "blocks ", 7) == 0) {
			binblocks = atoi(tmpbuf + 7);
		}else if(memcmp(tmpbuf, "name ", 5) == 0) {
			snprintf(BootHeader.title, 0x20, "%s", tmpbuf + 5);
			for(i = 0; i < 0x20; i++) {
				if((BootHeader.title[i] == 0x0A) || (BootHeader.title[i] == 0x0D))
					BootHeader.title[i] = 0;
			}
			BootHeader.title[0x1F] = 0;
		}else if(memcmp(tmpbuf, "makerid ", 8) == 0) {
			snprintf(BootHeader.maker_id, 4, "%s", tmpbuf + 8);
			for(i = 0; i < 4; i++) {
				if((BootHeader.maker_id[i] == 0x0A) || (BootHeader.maker_id[i] == 0x0D))
					BootHeader.maker_id[i] = 0;
			}
			BootHeader.maker_id[3] = 0;
		}else if(memcmp(tmpbuf, "maker ", 6) == 0) {
			snprintf(BootHeader.maker_name, 60, "%s", tmpbuf + 6);
			for(i = 0; i < 60; i++) {
				if((BootHeader.maker_name[i] == 0x0A) || (BootHeader.maker_name[i] == 0x0D))
					BootHeader.maker_name[i] = 0;
			}
			BootHeader.maker_name[59] = 0;
		}else if(memcmp(tmpbuf, "date ", 5) == 0) {
			snprintf(BootHeader.date, 8, "%s", tmpbuf + 5);
		}else if(memcmp(tmpbuf, "country ", 8) == 0) {
			BootHeader.country = atoi(tmpbuf + 8);
		}else if(memcmp(tmpbuf, "version ", 8) == 0) {
			BootHeader.version = atoi(tmpbuf + 8);
		}
	}
	fclose(fp);

	if(binname[0] == 0) {
		perror("No binary name");
		return EXIT_FAILURE;
	}

	if(!(in_fp = fopen(binname, "rb"))) {
		perror("Error opening input file");
		return EXIT_FAILURE;
	}

	obinname = malloc(strlen(argv[2])+5);
	sprintf(obinname, "%s.bin", argv[2]);
	if(!(out_fp = fopen(obinname, "wb"))) {
		perror("Error opening output file");
		return EXIT_FAILURE;
	}

	if(fstat(fileno(in_fp), &stat_buf)) {
		perror("fstat error");
		return EXIT_FAILURE;
	}

	if(binblocks == 0)
		binblocks = (stat_buf.st_size + 2047) / 2048;
	sector_count = binblocks;

	printf("Code+data Size: %ld\n", binblocks * 2048);
	sh_size = (1024 * 2048) - (binblocks * 2048) - 0x8000 - 2048;
	/* 2KiB minimum for stack or heap(they grow in different directions!) */

	if(sh_size < 0) {
		perror("Program size is too large");
		return EXIT_FAILURE;
	}
	printf("Stack+heap Free Space: %d\n", sh_size);

	BootHeader.sect_off = sector_offset;
	BootHeader.sect_count = sector_count;
	BootHeader.prog_off = prog_offset;
	BootHeader.prog_point = prog_point;

	BootHeader.sect_off = le32(BootHeader.sect_off);
	BootHeader.sect_count = le32(BootHeader.sect_count);
	BootHeader.prog_off = le32(BootHeader.prog_off);
	BootHeader.prog_point = le32(BootHeader.prog_point);
	BootHeader.volume_no = le32(BootHeader.volume_no);
	BootHeader.version = le16(BootHeader.version);
	BootHeader.country = le16(BootHeader.country);
	fwrite(&BootHeader, 1, sizeof(BootHeader), out_fp);
	padcnt += sizeof(BootHeader);

	data = malloc(stat_buf.st_size);
	if(fread(data, stat_buf.st_size, 1, in_fp) != 1) {
		perror("Can't read input data");
		return EXIT_FAILURE;
	}
	fclose(in_fp);
	fwrite(data, stat_buf.st_size, 1, out_fp);
	padcnt += stat_buf.st_size;

	while((padcnt & 2047) || (padcnt < (2048 * 75 * 4))) {
		fputc(0, out_fp);
		padcnt++;
	}
	fclose(out_fp);

	/* Output a simple .cue file. */
	cuename = malloc(strlen(argv[2])+5);
	sprintf(cuename, "%s.cue", argv[2]);
	fp = fopen(cuename, "wb");
	fprintf(fp, "FILE \"%s\" BINARY\n", obinname);
	fprintf(fp, "  TRACK 01 MODE1/2048\n");
	fprintf(fp, "    INDEX 01 00:00:00\n");
	fclose(fp);
	free(cuename);
	free(obinname);

	printf("Done.\n");
	return EXIT_SUCCESS;
}
